--[[
    - "The colon is for implementing methods that pass self as the first parameter. So x:bar(3,4)should be the same as x.bar(x,3,4)."
      https://stackoverflow.com/questions/4911186/difference-between-and-in-lua

    - _G[] access the global namespace
      https://stackoverflow.com/questions/35910099/how-special-is-the-global-variable-g
]]

TheMotherTongue = {} -- global table to access the addon
local TMT = TheMotherTongue -- local name for ease of work
TMT.version = "v9.2.0_2.0"

--[[
    If some addon wants to change the language of the names displayed, these are
    the tables that needs to be changed.
    It has to be collections of the form : {{id, name}, {id, name}, ...}
]]
TMT.npcs = {}
TMT.spells = {}
TMT.items = {}
TMT.quests = {}
TMT.areas = {}

local TMT_AvailableLanguages = {}
local TMT_NoLanguagePack = false
local TMT_Languages = {
    en = "English",
    fr = "Français",
    de = "Deutsch",
    es = "Español",
    pt = "Português",
    it = "Italiano",
    ru = "pусский",
    ko = "한국어",
    cn = "简体中文",
}

-- default values of settings that the user can change
local TMT_Defaults = {
    language = "",
    npcNameColor = {255, 128, 0}, -- name color in tooltip
    questNameColor = {255, 106, 0}, -- name color in tooltip
    spellNameColor = {0, 128, 255}, -- name color in tooltip
    itemNameColor = {87, 152, 255}, -- name color in tooltip
    displayNPCsNames = true,
    npcTooltip_DisplayNameInTitle = false,
    displayQuestsNames = true,
    displaySpellsNames = true,
    spellTooltip_DisplayNameInTitle = false,
    displayItemsNames = true,
    itemTooltip_DisplayNameInTitle = true,
    itemTooltip_UseBlizzardsColor = true,
    questsNames_UseBlizzardsColor = true,
    spellTooltip_UseClassColor = true,
    tooltips_DisplayNameUnderTitle = false,
    questTooltip_DisplayNameInTitle = true,
    npcTooltip_UseBlizzardsColor = false,
    lastUpdateMessageVersion = "",
}

--[[
    A word on colors in TMT :
    To display text in color in Blizzard UI easily, we have to use the hex form,
    preceded by "|cFF".
    User colors preferences are stored in RGB. It's because the color picker works
    in rgb and not in hexadecimal.
    Oh, and rgb in color picker is between 0 and 1.
]]
local TMT_Colors = {
    orange = "|cFFFF8000",
    red = "|cFFFF0000",
    blue = "|cFF0070dd",
    magenta = "|cFFFF0040",
    grey = "|cFF191203",
    lightGrey = "|cFFB3B3B3",
    brown = "|cFF5E130B",
    lightBrown = "|cFFAC240B",
    green = "|cFF049713",
    white = "|cFFFFFFFF",
    yellow = "|cFFFFD100",
    classes = {
        WARRIOR = "|cFFC79C6E",
        PALADIN = "|cFFF58CBA",
        HUNTER = "|cFFA9D271",
        ROGUE = "|cFFFFF569",
        PRIEST = "|cFFFFFFFF",
        DEATHKNIGHT = "|cFFC41F3B",
        SHAMAN = "|cFF0070DE",
        MAGE = "|cFF40C7EB",
        WARLOCK = "|cFF8787ED",
        MONK = "|cFF00FF96",
        DRUID = "|cFFFF7D0A",
        DEMONHUNTER = "|cFFA330C9"
    },
    worldQuestQuality = {"|cFFFFFFFF", "|cFF0070dd", "|cFFa335ee"} -- white, blue, purple
}

local TMT_ZoneType = {
    ZONE = 1,
    SUBZONE = 2
}

--[[
    Button to test for text width more accurately. So instead of test string length,
    we set the button text, and then test the fontString width
]] 
local TMT_DebugButton = CreateFrame("Button", TMT_QuestsFrame_GlobalName)
local TMT_VMargin = 16
local TMT_HMargin = 15
local TMT_Tab = "      "
local TMT_ColorBeingChanged -- used by the color picker to know which color to change
local TMT_SpellsMaxId = 400000
local TMT_QuestsMaxId = 200000
local TMT_AreasMaxId = 14000
local TMT_ItemsMaxId = 185000
local TMT_EntriesPerFile = 25000
local TMT_Talents = {}
local TMT_PvpTalents = {}

local function TMT_WriteInChat(something)
    local chatFrame = DEFAULT_CHAT_FRAME.editBox
    ChatEdit_ActivateChat(chatFrame)
    chatFrame:SetText(chatFrame:GetText() .. something)
end

local function TMT_GetItemsNames()
    TMT_NamesDump = {}
    local pair = {}
    local name

    for i = 0, TMT_ItemsMaxId do
        name = C_Item.GetItemNameByID(i)
        if name then
            pair = {i, name}
            table.insert(TMT_NamesDump, pair)
        end
    end

    print(TMT_Colors.blue .. "Done")
end

--Gets all spells names with their id
local function TMT_GetSpellsNames()
    TMT_NamesDump = {}
    local pair = {}
    local name, rank, icon, castTime, minRange, maxRange

    for i = 346900, TMT_SpellsMaxId do
        name, rank, icon, castTime, minRange, maxRange = GetSpellInfo(i)
        if name then
            pair = {i, name}
            table.insert(TMT_NamesDump, pair)
        end
    end

    print(TMT_Colors.blue .. "Done")
end

--Gets all areas names with their id
local function TMT_GetAreas()
    TMT_NamesDump = {}
    local pair = {}
    local name

    for i = 1, TMT_AreasMaxId do
        name = C_Map.GetAreaInfo(i)
        if name then
            pair = {i, name}
            table.insert(TMT_NamesDump, pair)
        end
    end

    print(TMT_Colors.blue .. "Done")
end

local function TMT_GetWorldQuestsNames()
    TMT_NamesDump = {}
    local pair = {}
    local name

    for i = 1, TMT_QuestsMaxId do
        name = C_TaskQuest.GetQuestInfoByQuestID(i)
        if name then
            pair = {i, name}
            table.insert(TMT_NamesDump, pair)
        end
    end

    print(TMT_Colors.blue .. "Done")
end

local function TMT_GetStringWidth(text, font)
    if font then
        TMT_DebugButton:SetNormalFontObject(font)
    else
        TMT_DebugButton:SetNormalFontObject("QuestFontNormalSmall")
    end

    TMT_DebugButton:SetText(text)
    return math.floor(TMT_DebugButton:GetFontString():GetStringWidth() + 0.5)
end

local function TMT_LanguageDropdown(settings)
    local values = settings['values'] or {}
    local level = 1
    local changeHandler = settings['changeHandler']
    local dropdown = CreateFrame("Frame", settings['name'], settings['parent'], 'UIDropDownMenuTemplate')

    -- Sets the dropdown width
    local dropdownWidth = 0
    for _, value in pairs(values) do 
        local textWidth = TMT_GetStringWidth(value, "GameFontNormal") + 30
        if textWidth > dropdownWidth then
            dropdownWidth = textWidth
        end
    end

    UIDropDownMenu_SetWidth(dropdown, dropdownWidth)
    UIDropDownMenu_SetText(dropdown, TMT_Languages[TMT_Language])

    UIDropDownMenu_Initialize(dropdown, function(self, level, _)
        local info = UIDropDownMenu_CreateInfo()

        for key, value in pairs(values) do
            info.text = value
            info.checked = key == TMT_Language
            info.menuList = key
            info.hasArrow = false
            info.arg1 = key
            info.func = function(info, arg1, arg2, checked)
                info.checked = true
                UIDropDownMenu_SetText(dropdown, info.value)
                changeHandler(dropdown, info)
            end
            UIDropDownMenu_AddButton(info)
        end
    end)

    return dropdown
end

local function TMT_SetLanguage() 
    TMT.npcs = _G["TMT_Npcs_" .. TMT_Language]
    TMT.spells = _G["TMT_Spells_" .. TMT_Language]
    TMT.items = _G["TMT_Items_" .. TMT_Language]
    TMT.quests = _G["TMT_Quests_" .. TMT_Language]
    TMT.talents = _G["TMT_Talents_" .. TMT_Language]
    TMT.areas = _G["TMT_Areas_" .. TMT_Language] 
    --TMT.instances = _G["TMT_Instances_" .. TMT_Language]
    --TMT.maps = _G["TMT_Maps_" .. TMT_Language]

    --removes the other languages tables from memory
    for key,val in pairs(TMT_AvailableLanguages) do
        if key ~= TMT_Language then
            _G["TMT_Npcs_" .. key] = nil
            _G["TMT_Spells_" .. key] = nil
            _G["TMT_Items_" .. key] = nil
            _G["TMT_Quests_" .. key] = nil
            _G["TMT_Talents_" .. key] = nil
            _G["TMT_Areas_" .. key] = nil 
            --_G["TMT_Instances_" .. key] = nil
            --_G["TMT_Maps_" .. key] = nil
        end
    end
end

local function TMT_RgbToHex(rgb)
    local hexadecimal = ""
    local digitIndex

    local r, g, b = unpack(rgb)
    local value = r * 65536 + g * 256 + b

    if value == 0 then
        hexadecimal = "000000"
    else     
        while value > 0 do
            digitIndex = math.fmod(value, 16) + 1
            value = math.floor(value / 16)
            hexadecimal = string.sub('0123456789ABCDEF', digitIndex, digitIndex) .. hexadecimal	
        end
    end

    while string.len(hexadecimal) < 6 do
        hexadecimal = "0" .. hexadecimal
    end

	return "|cff" .. hexadecimal
end

local function TMT_SetButtonState(param, button)
    if param then
        button:Enable()
    else
        button:Disable()
    end
end

local function TMT_ChangeFramePosition(frame, anchor, parent, parentAnchor, x, y)
    x = x or 0
    y = y or 0
    
    frame:ClearAllPoints()
    frame:SetPoint(anchor, parent, parentAnchor, x, y)
end

-- Used to be sure the config panel is always matching saved variables
local function TMT_ConfigPanelSanityCheck()
    if not TMT_NoLanguagePack then
        TMT_NpcsNamesCheckBox_GlobalName:SetChecked(TMT_DisplayNPCsNames)
        TMT_NpcsInTitleCheckBox_GlobalName:SetChecked(TMT_NPCTooltip_DisplayNameInTitle)
        TMT_NpcsBlizzardColorCheckButton_GlobalName:SetChecked(TMT_NPCTooltip_UseBlizzardsColor)

        TMT_SpellsNamesCheckBox_GlobalName:SetChecked(TMT_DisplaySpellsNames)
        TMT_SpellsInTitleCheckBox_GlobalName:SetChecked(TMT_SpellTooltip_DisplayNameInTitle)
        TMT_SpellsClassColorCheckBox_GlobalName:SetChecked(TMT_SpellTooltip_UseClassColor)

        TMT_ItemsNamesCheckBox_GlobalName:SetChecked(TMT_DisplayItemsNames)
        TMT_ItemsInTitleCheckBox_GlobalName:SetChecked(TMT_ItemTooltip_DisplayNameInTitle)
        TMT_ItemsBlizzardColorCheckButton_GlobalName:SetChecked(TMT_ItemTooltip_UseBlizzardsColor)

        TMT_QuestsNamesCheckButton_GlobalName:SetChecked(TMT_DisplayQuestsNames)
        TMT_QuestsBlizzardColorCheckButton_GlobalName:SetChecked(TMT_QuestsNames_UseBlizzardsColor)
        TMT_QuestsInTitleCheckBox_GlobalName:SetChecked(TMT_QuestTooltip_DisplayNameInTitle)

        TMT_NamesUnderTitleCheckBox_GlobalName:SetChecked(TMT_Tooltips_DisplayNameUnderTitle)

        TMT_SetButtonState(TMT_DisplayNPCsNames, TMT_NpcsInTitleCheckBox_GlobalName)
        TMT_SetButtonState(not TMT_NPCTooltip_UseBlizzardsColor and TMT_DisplayNPCsNames, TMT_NPCsColorButton_GlobalName)
        TMT_SetButtonState(TMT_DisplayNPCsNames, TMT_NpcsBlizzardColorCheckButton_GlobalName)

        TMT_SetButtonState(TMT_DisplaySpellsNames, TMT_SpellsInTitleCheckBox_GlobalName)
        TMT_SetButtonState(TMT_DisplaySpellsNames, TMT_SpellsClassColorCheckBox_GlobalName)
        TMT_SetButtonState(not TMT_SpellTooltip_UseClassColor and TMT_DisplaySpellsNames, TMT_SpellsColorButton_GlobalName)

        TMT_SetButtonState(TMT_DisplayItemsNames, TMT_ItemsInTitleCheckBox_GlobalName)
        TMT_SetButtonState(TMT_DisplayItemsNames, TMT_ItemsBlizzardColorCheckButton_GlobalName)
        TMT_SetButtonState(not TMT_ItemTooltip_UseBlizzardsColor and TMT_DisplayItemsNames, TMT_ItemsColorButton_GlobalName)
        
        TMT_SetButtonState(TMT_DisplayQuestsNames and not TMT_QuestsNames_UseBlizzardsColor, TMT_QuestsColorButton_GlobalName)
        TMT_SetButtonState(TMT_DisplayQuestsNames, TMT_QuestsInTitleCheckBox_GlobalName)
        TMT_SetButtonState(TMT_DisplayQuestsNames, TMT_QuestsBlizzardColorCheckButton_GlobalName)
    end
end

local function TMT_RestoreDefaults_Quests()
    TMT_QuestNameColor = TMT_Defaults.questNameColor
    TMT_DisplayQuestsNames = TMT_Defaults.displayQuestsNames
    TMT_QuestTooltip_DisplayNameInTitle = TMT_Defaults.questTooltip_DisplayNameInTitle
    TMT_QuestsNames_UseBlizzardsColor = TMT_Defaults.questsNames_UseBlizzardsColor

    TMT_ConfigPanelSanityCheck()
end

local function TMT_RestoreDefaults_NPCs()
    TMT_NpcNameColor = TMT_Defaults.npcNameColor
    TMT_DisplayNPCsNames = TMT_Defaults.displayNPCsNames
    TMT_NPCTooltip_DisplayNameInTitle = TMT_Defaults.npcTooltip_DisplayNameInTitle
    TMT_NPCTooltip_UseBlizzardsColor = TMT_Defaults.npcTooltip_UseBlizzardsColor

    TMT_ConfigPanelSanityCheck()
end

local function TMT_RestoreDefaults_Items()
    TMT_ItemNameColor = TMT_Defaults.itemNameColor
    TMT_ItemTooltip_DisplayNameInTitle = TMT_Defaults.itemTooltip_DisplayNameInTitle
    TMT_ItemTooltip_UseBlizzardsColor = TMT_Defaults.itemTooltip_UseBlizzardsColor
    TMT_DisplayItemsNames = TMT_Defaults.displayItemsNames

    TMT_ConfigPanelSanityCheck()
end

local function TMT_RestoreDefaults_Spells()
    TMT_SpellNameColor = TMT_Defaults.spellNameColor
    TMT_DisplaySpellsNames = TMT_Defaults.displaySpellsNames
    TMT_SpellTooltip_DisplayNameInTitle = TMT_Defaults.spellTooltip_DisplayNameInTitle
    TMT_SpellTooltip_UseClassColor = TMT_Defaults.spellTooltip_UseClassColor

    TMT_ConfigPanelSanityCheck()
end

local function TMT_RestoreDefaults_Globals()
    TMT_Tooltips_DisplayNameUnderTitle = TMT_Defaults.tooltips_DisplayNameUnderTitle
end

local function TMT_RestoreDefaults()
    TMT_RestoreDefaults_Globals()
    TMT_RestoreDefaults_Spells()
    TMT_RestoreDefaults_NPCs()
    TMT_RestoreDefaults_Items()
    TMT_RestoreDefaults_Quests()

    TMT_ConfigPanelSanityCheck()
end

local function TMT_CheckboxTemplate(globalName, text, tooltip, hMargin, vMargin, parent)
    local checkBox = CreateFrame("CheckButton", globalName, parent, "ChatConfigCheckButtonTemplate")
    _G[checkBox:GetName() .. "Text"]:SetText(text);
    checkBox.tooltip = tooltip
    checkBox:SetPoint("TOPLEFT", hMargin, vMargin)

    return checkBox
end

local function TMT_DefaultButton(parent, relativeAnchor, anchor, width, height, hMargin, vMargin, text, font, globalName)
    local button = CreateFrame("Button", globalName, parent, "UIPanelButtonTemplate")
    button:SetWidth(width)
    button:SetHeight(height)
    button:SetPoint(anchor, parent, relativeAnchor, hMargin, vMargin)

    local buttonText = button:CreateFontString(nil, "ARTWORK", font)
    buttonText:SetPoint("CENTER", button, "CENTER")
    buttonText:SetText(text)

    return button
end

--[[
    This function is called when :
        - "ok" or "cancel" is clicked in the color picker frame
        - the color is changed in the color picker frame
        - the opacity is changed in the color picker frame (not handled here)

    previousColor is set when user clicks on "cancel", nil otherwise
]]
local function TMT_ColorCallBack(previousColor)
    local r, g, b, a

    if previousColor then
        r, g, b, a = unpack(previousColor)
    else
        r, g, b = ColorPickerFrame:GetColorRGB()
        a = 1 - OpacitySliderFrame:GetValue() -- slider values are reversed (0 at the top), weird
    end

    if TMT_ColorBeingChanged == "NPC" then 
        -- color picker works between 0 .. 1, TMT works between 0 .. 255
        TMT_NpcNameColor = {r * 255, g * 255, b * 255} 
    elseif TMT_ColorBeingChanged == "spell" then  
        TMT_SpellNameColor = {r * 255, g * 255, b * 255}
    elseif TMT_ColorBeingChanged == "item" then
        TMT_ItemNameColor = {r * 255, g * 255, b * 255}
    elseif TMT_ColorBeingChanged == "quest" then
        TMT_QuestNameColor = {r * 255, g * 255, b * 255}
    end
end

-- Called when color picker is shown
local function TMT_ShowColorPicker(rgba, colorBeingChanged, callBack)
    local r,g,b,a = unpack(rgba)
    -- TMT works between 0 and 255, color Picker works between 0 and 1
    r = r / 255
    g = g / 255
    b = b / 255

    TMT_ColorBeingChanged = colorBeingChanged

    ColorPickerFrame.func, ColorPickerFrame.opacityFunc, ColorPickerFrame.cancelFunc = callBack, callBack, callBack
    ColorPickerFrame:SetColorRGB(r,g,b)
    ColorPickerFrame.hasOpacity = a ~= nil

    if ColorPickerFrame.hasOpacity then
        ColorPickerFrame.opacity = 1 - a -- slider values are reversed (0 at the top), weird
    end
    
    ColorPickerFrame.previousValues = {r,g,b,a}

    -- apparently needed to correctly throw "show" event
    ColorPickerFrame:Hide()
    ColorPickerFrame:Show()
end

local function TMT_GetClassColor()
    local localeClass, englishClass, classIndex = UnitClass("player")

    return TMT_Colors.classes[englishClass]
end

local function TMT_LanguageChange(dropdown, info)
    local previousLanguage = TMT_Language
    TMT_Language = info.arg1

    if previousLanguage ~= TMT_Language then
        --[[
            We have to reload to get all languages tables back since we delete 
            them to save memory when they are not the active language tables
        ]]
        ReloadUI()
        TMT_SetLanguage()
    end

    TMT_ConfigPanelSanityCheck()
end

-- Configuration panel
local function TMT_CreateConfigPanel()
    local configPanel = CreateFrame("FRAME", "TMT_ConfigPanel_GlobalName", UIParent)
    local lineWidth = 136
    local columnWidth = 180
    local firstColumn = TMT_HMargin
    local secondColumn = firstColumn + columnWidth
    local thirdColumn = secondColumn + columnWidth
    local firstLine = -TMT_VMargin
    local secondLine = firstLine - lineWidth
    local thirdLine = secondLine - lineWidth
    local fourthLine = thirdLine - lineWidth
    local checkBoxSpace = 24
    local buttonSpace = 16
    local bigButtonSpace = 32
    local titleSpace = 32
    local textLineSpace = 16
    local tab = 15
    local HPos
    local VPos
    configPanel.name = "The Mother Tongue" -- The name in the "Add-ons" tab of interface
    InterfaceOptions_AddCategory(configPanel)

    -- Title
    HPos = firstColumn
    VPos = firstLine
    local configPanelTitle = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
    configPanelTitle:SetPoint("TOPLEFT", HPos, VPos)
    configPanelTitle:SetText("The Mother Tongue")
    
    -- Version
    HPos = firstColumn
    VPos = firstLine - textLineSpace
    local configPanelSubtitle = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
    configPanelSubtitle:SetPoint("TOPLEFT", HPos, VPos)
    configPanelSubtitle:SetText(TMT.version)

    if TMT_NoLanguagePack then
        HPos = 0
        VPos = secondLine
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
        globalCat:SetPoint("TOP", HPos, VPos)
        globalCat:SetText("You have no TMT language pack loaded !")

        HPos = 0
        VPos = secondLine - titleSpace
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
        globalCat:SetPoint("TOP", HPos, VPos)
        globalCat:SetText("You can get them like any other AddOn.")


        HPos = 0
        VPos = secondLine - titleSpace - textLineSpace - textLineSpace - textLineSpace
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
        globalCat:SetPoint("TOP", HPos, VPos)
        globalCat:SetText("Available packs : ")

        HPos = 0
        VPos = secondLine - titleSpace - textLineSpace - textLineSpace - textLineSpace - textLineSpace - textLineSpace
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
        globalCat:SetPoint("TOP", HPos, VPos)
        globalCat:SetText("TheMotherTongue_LanguagePack_en (English)")
    else
        --language selection title
        HPos = firstColumn - tab - tab
        VPos = firstLine
        local languageSelectionTitle = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        languageSelectionTitle:SetPoint("TOPRIGHT", HPos, VPos)
        languageSelectionTitle:SetText("Language")

        -- Language selection dropdown
        HPos = firstColumn - tab
        VPos = firstLine - textLineSpace
        local settings = {}
        settings.parent = configPanel
        settings.name = "TMT_LanguageSelectionDropdown_GlobalName"
        settings.values = TMT_AvailableLanguages
        settings.changeHandler = TMT_LanguageChange

        local dropdown = TMT_LanguageDropdown(settings)
        dropdown:SetPoint("TOPRIGHT", HPos, VPos)

        -- Reset Settings button
        HPos = firstColumn + tab
        VPos = firstLine - titleSpace
        local defaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 128, 24, HPos, VPos, "Reset Settings", "GameFontNormal")
        defaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults() end)

        -- Global parameters title
        HPos = firstColumn
        VPos = firstLine - titleSpace - bigButtonSpace - textLineSpace
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        globalCat:SetPoint("TOPLEFT", HPos, VPos)
        globalCat:SetText("Global settings")

        -- Globals parameters
        HPos = firstColumn + tab
        VPos = firstLine - titleSpace - bigButtonSpace - textLineSpace - textLineSpace
        local namesUnderTitleCheckBox = TMT_CheckboxTemplate("TMT_NamesUnderTitleCheckBox_GlobalName", " Display names under title",  "If checked, names set to be displayed in the title are displayed just under it instead", HPos, VPos, configPanel)
        namesUnderTitleCheckBox:SetScript("OnClick", function(self) TMT_Tooltips_DisplayNameUnderTitle = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- NPCs subcategory title
        HPos = firstColumn
        VPos = secondLine
        local npcCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        npcCat:SetPoint("TOPLEFT", HPos, VPos)
        npcCat:SetText("NPCs settings")

        -- Reset settings button
        HPos = firstColumn + tab * 6
        VPos = secondLine
        local npcsDefaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 48, 16, HPos, VPos, "Reset", "GameFontNormalSmall")
        npcsDefaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults_NPCs() end)

        -- NPCs names checkbox
        HPos = firstColumn + tab
        VPos = secondLine - textLineSpace
        local npcsNamesCheckBox = TMT_CheckboxTemplate("TMT_NpcsNamesCheckBox_GlobalName", " NPCs names",  "If checked, displays the English name of the NPC the mouse is hovering above in the tooltip", HPos, VPos, configPanel)
        npcsNamesCheckBox:SetScript("OnClick", function(self) TMT_DisplayNPCsNames = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)
        
        -- NPCs in title checkbox
        HPos = firstColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace
        local npcsInTitleCheckBox = TMT_CheckboxTemplate("TMT_NpcsInTitleCheckBox_GlobalName", " In title", "If checked, displays the name in the tooltip title. If not, displays it on a new line at the end of the tooltip.", HPos, VPos, configPanel)
        npcsInTitleCheckBox:SetScript("OnClick", function(self) TMT_NPCTooltip_DisplayNameInTitle = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- NPCs Blizz's color button
        HPos = firstColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace
        local npcsBlizzardColorCheckButton = TMT_CheckboxTemplate("TMT_NpcsBlizzardColorCheckButton_GlobalName", " Use Blizzard's colors", "If checked, uses Blizzard's colors to display NPCs English names", HPos, VPos, configPanel)
        npcsBlizzardColorCheckButton:SetScript("OnClick", function(self) TMT_NPCTooltip_UseBlizzardsColor = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- NPCs color button
        HPos = firstColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace - checkBoxSpace
        local npcsColorButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 80, 19, HPos, VPos, "Name Color", "GameFontNormalSmall", "TMT_NPCsColorButton_GlobalName")
        npcsColorButton:SetScript("OnClick", function(self) TMT_ShowColorPicker(TMT_NpcNameColor, "NPC", TMT_ColorCallBack) end)

        -- Spells subcategory title
        HPos = secondColumn
        VPos = secondLine
        local spellsCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        spellsCat:SetPoint("TOPLEFT", HPos, VPos)
        spellsCat:SetText("Spells settings")

        -- Reset settings button
        HPos = secondColumn + tab * 6
        VPos = secondLine
        local spellsDefaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 48, 16, HPos, VPos, "Reset", "GameFontNormalSmall")
        spellsDefaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults_Spells() end)

        -- Spells names checkbox
        HPos = secondColumn + tab
        VPos = secondLine - textLineSpace
        local spellsNamesCheckBox = TMT_CheckboxTemplate("TMT_SpellsNamesCheckBox_GlobalName", " Spells names", "If checked, displays the English name of the spell the mouse is hovering above in the tooltip", HPos, VPos, configPanel)
        spellsNamesCheckBox:SetScript("OnClick", function(self) TMT_DisplaySpellsNames = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Spells in title checkbox
        HPos = secondColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace
        local spellsInTitleCheckBox = TMT_CheckboxTemplate("TMT_SpellsInTitleCheckBox_GlobalName", " In title", "If checked, displays the name in the tooltip title. If not, displays it on a new line at the end of the tooltip.", HPos, VPos, configPanel)
        spellsInTitleCheckBox:SetScript("OnClick", function(self) TMT_SpellTooltip_DisplayNameInTitle = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Spells class color checkBox
        HPos = secondColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace
        local spellsClassColorCheckBox = TMT_CheckboxTemplate("TMT_SpellsClassColorCheckBox_GlobalName", " Use class color", "If checked, uses your class color", HPos, VPos, configPanel)
        spellsClassColorCheckBox:SetScript("OnClick", function(self) TMT_SpellTooltip_UseClassColor = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Spells color button
        HPos = secondColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace - checkBoxSpace
        local spellsColorButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 80, 19, HPos, VPos, "Name Color", "GameFontNormalSmall", "TMT_SpellsColorButton_GlobalName")
        spellsColorButton:SetScript("OnClick", function(self) TMT_ShowColorPicker(TMT_SpellNameColor, "spell", TMT_ColorCallBack) end)

        -- Items subcategory title
        HPos = thirdColumn
        VPos = secondLine
        local itemsCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        itemsCat:SetPoint("TOPLEFT", HPos, VPos)
        itemsCat:SetText("Items settings")

        -- Reset settings button
        HPos = thirdColumn + tab * 6
        VPos = secondLine
        local itemsDefaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 48, 16, HPos, VPos, "Reset", "GameFontNormalSmall")
        itemsDefaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults_Items() end)

        -- Items names checkbox
        HPos = thirdColumn + tab
        VPos = secondLine - textLineSpace
        local itemsNamesCheckBox = TMT_CheckboxTemplate("TMT_ItemsNamesCheckBox_GlobalName", " Items names", "If checked, displays the English name of the item the mouse is hovering above in the tooltip", HPos, VPos, configPanel)
        itemsNamesCheckBox:SetScript("OnClick", function(self) TMT_DisplayItemsNames = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Items in title checkbox
        HPos = thirdColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace
        local itemsInTitleCheckBox = TMT_CheckboxTemplate("TMT_ItemsInTitleCheckBox_GlobalName", " In title", "If checked, displays the name in the tooltip title. If not, displays it on a new line at the end of the tooltip.", HPos, VPos, configPanel)
        itemsInTitleCheckBox:SetScript("OnClick", function(self) TMT_ItemTooltip_DisplayNameInTitle = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Items tooltip blizzard's color checkbox
        HPos = thirdColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace
        local itemsBlizzardColorCheckButton = TMT_CheckboxTemplate("TMT_ItemsBlizzardColorCheckButton_GlobalName", " Use Blizzard's colors", "If checked, uses Blizzard's colors to display items English names", HPos, VPos, configPanel)
        itemsBlizzardColorCheckButton:SetScript("OnClick", function(self) TMT_ItemTooltip_UseBlizzardsColor = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Items color button
        HPos = thirdColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace - checkBoxSpace
        local itemsColorButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 80, 19, HPos, VPos, "Name Color", "GameFontNormalSmall", "TMT_ItemsColorButton_GlobalName")
        itemsColorButton:SetScript("OnClick", function(self) TMT_ShowColorPicker(TMT_ItemNameColor, "item", TMT_ColorCallBack) end)

        -- Quests subcategory title
        HPos = firstColumn
        VPos = thirdLine
        local questsCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        questsCat:SetPoint("TOPLEFT", HPos, VPos)
        questsCat:SetText("Quests settings")

        -- Reset settings button
        HPos = firstColumn + tab * 6
        VPos = thirdLine
        local questsDefaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 48, 16, HPos, VPos, "Reset", "GameFontNormalSmall")
        questsDefaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults_Quests() end)

        -- Quests names checkbox
        HPos = firstColumn + tab
        VPos = thirdLine - textLineSpace
        local questCheckButton = TMT_CheckboxTemplate("TMT_QuestsNamesCheckButton_GlobalName", " Quests names", "If checked, displays the English name of the quests currently in your questlog", HPos, VPos, configPanel)
        questCheckButton:SetScript("OnClick", function(self) TMT_DisplayQuestsNames = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Quests in title checkbox
        HPos = firstColumn + tab
        VPos = thirdLine - textLineSpace - checkBoxSpace
        local questsInTitleCheckBox = TMT_CheckboxTemplate("TMT_QuestsInTitleCheckBox_GlobalName", " In title", "If checked, displays the name in the tooltip title. If not, displays it on a new line at the end of the tooltip.", HPos, VPos, configPanel)
        questsInTitleCheckBox:SetScript("OnClick", function(self) TMT_QuestTooltip_DisplayNameInTitle = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Quests Blizz's colors checkbox
        HPos = firstColumn + tab
        VPos = thirdLine - textLineSpace - checkBoxSpace - checkBoxSpace
        local questsBlizzardColorCheckButton = TMT_CheckboxTemplate("TMT_QuestsBlizzardColorCheckButton_GlobalName", " Blizzard's colors", "If checked, uses Blizzard's colors to display quests English names", HPos, VPos, configPanel)
        questsBlizzardColorCheckButton:SetScript("OnClick", function(self) TMT_QuestsNames_UseBlizzardsColor = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Quests color button
        HPos = firstColumn + tab
        VPos = thirdLine - textLineSpace - checkBoxSpace - checkBoxSpace - checkBoxSpace
        local questsColorButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 80, 19, HPos, VPos, "Name Color", "GameFontNormalSmall", "TMT_QuestsColorButton_GlobalName")
        questsColorButton:SetScript("OnClick", function(self) TMT_ShowColorPicker(TMT_QuestNameColor, "quest", TMT_ColorCallBack) end)
    
       -- Notes
       HPos = firstColumn
       VPos = fourthLine - checkBoxSpace
       local notesTitle = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
       notesTitle:SetPoint("TOPLEFT", HPos, VPos)
       notesTitle:SetText("Notes")

       HPos = firstColumn + tab
       VPos = fourthLine - checkBoxSpace - textLineSpace
       local notes1 = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
       notes1:SetPoint("TOPLEFT", HPos, VPos)
       notes1:SetText("- You can get a TMT language pack like any other AddOn. Available packs : ")

       HPos = firstColumn + tab + tab
       VPos = fourthLine - checkBoxSpace - textLineSpace - textLineSpace
       local notes2 = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
       notes2:SetPoint("TOPLEFT", HPos, VPos)
       notes2:SetText("- TheMotherTongue_LanguagePack_en (English)")
    end

    TMT_ConfigPanelSanityCheck()
end

-- Slash commands code
local function TMT_Commands(msg, editbox)
    if msg == "version" or msg == "v" then
        print(TMT_Colors.orange .. "TMT " .. TMT_Colors.green .. TMT.version)
    elseif msg == "getspellsnames" then
        TMT_GetSpellsNames()
    elseif msg == "getworldquestsnames" then
        TMT_GetWorldQuestsNames()
    elseif msg == "getareasnames" then
        TMT_GetAreas()
    elseif msg == "getitemsnames" then
        TMT_GetItemsNames()
    else 
        -- called twice to work around a BLizzard's "bug"
        InterfaceOptionsFrame_OpenToCategory(TMT_ConfigPanel_GlobalName)
        InterfaceOptionsFrame_OpenToCategory(TMT_ConfigPanel_GlobalName)
    end

    TMT_ConfigPanelSanityCheck()
end

SLASH_TMT1 = "/tmt"
SLASH_TMT2 = "/themothertongue"
SlashCmdList["TMT"] = TMT_Commands

local function TMT_DetectLanguagesPacks()
    local packageFound = false
    local firstPackageFound

    for key,val in pairs(TMT_Languages) do
        local language = _G["TheMotherTongue_LanguagePack_" .. key];
        if language then
            if(not packageFound) then
                firstPackageFound = key
            end
            packageFound = true
            TMT_AvailableLanguages[key] = val
        end
    end
    
    -- Sets the active language
    if packageFound then
        if not TMT_AvailableLanguages[TMT_Language] then
            -- We had a language pack before, but it's no longer there
            if TMT_AvailableLanguages.en then
                TMT_Language = "en"
            else
                TMT_Language = firstPackageFound
            end
        end

        TMT_SetLanguage()
    else
        TMT_Language = ""
        print(TMT_Colors.orange .. "TMT: " .. TMT_Colors.red 
            .. "No TMT language pack detected ! Type '/tmt' to learn how to get them")
    end

    -- Deactivate / activate the displays
    TMT_DisplayNPCsNames = packageFound
    TMT_DisplayQuestsNames = packageFound
    TMT_DisplaySpellsNames = packageFound
    TMT_DisplayItemsNames = packageFound
    TMT_DisplayAreasNames = packageFound
    TMT_NoLanguagePack = not packageFound
end

local function TMT_DisplayUpdateMessage()
    local updateMessage = TMT_Colors.orange .. "TMT: Language packs are now supported. Don't forget to install the english one ! Type '/tmt' to learn how to get it. (Coming soon : German pack)." 

    if TMT.version ~= TMT_LastUpdateMessageVersion then
        print(updateMessage)
        TMT_LastUpdateMessageVersion = TMT.version
    end
end

-- idNamePairs is of the form {id, "name"}
local function TMT_FindNameFromId(idToFind, idNamePairs)
    local name

    for _, pair in ipairs(idNamePairs) do
        if pair[1] == idToFind then
            name = pair[2]
            break
        end
    end

    return name
end

--[[
    Sometimes the hooked scripts are called more than once by the API. This does happen
    for talents, recipes, formulas, patrons.
    This function is used to check if the name is already in the tooltip to not display it
    more than once.
    It's used only if I can't find a way to make the calls happen only once.
]]
local function TMT_IsNameAlreadyInTooltip(tooltip, name)
    local found = false

    if name then
        local tooltipName = tooltip:GetName()

        for line = 1, tooltip:NumLines() do
            if string.match(_G[tooltipName .. "TextLeft" .. line]:GetText(), name) then
                found = true
                break
            end
        end
    end

    return found
end

local function TMT_TooltipDisplay(tooltip, name, inTitle, color, isArea)
    isArea = isArea or false

    if name then
        if not color then
            color = TMT_Colors.orange
        end
        
        -- Doesn't display anything if the name in English is the same
        if GameTooltipTextLeft1:GetText() ~= name then
            if isArea then 
                local separator = TMT_Colors.lightGrey .. ""
                tooltip:AddLine(separator)
                tooltip:AddLine(color .. name)
            else
                if inTitle then
                    local placement = "|r" .. TMT_Colors.lightGrey .. " - "
                    
                    if TMT_Tooltips_DisplayNameUnderTitle then
                        placement = "|n"
                    end

                    tooltip:AppendText(placement .. color .. name)
                else
                    tooltip:AddLine(color .. name)
                end
            end
        end
    end
end

--[[
    Parse the GUID to get the actual ID of the unit.
    A unit GUID is of the form : Creature-0-4480-1-244-2955-0006FF9B5.
    The ID we want here would be 2955 (a Plainstrider btw).
]]
local function TMT_GetUnitId(guid)
    if guid:match("^%a+") ~= "Player" then
        return tonumber(guid:match("-(%d+)-%x+$"), 10)
    end
end

local function TMT_FindNpcOriginalName(unit, fromGuid)
    local unitName

    if unit then
        local npcId = nil

        if fromGuid then
            local guid = UnitGUID(unit)
            npcId = TMT_GetUnitId(guid)
        else
            npcId = unit
        end

        if npcId then
            local npcsArray = TMT.npcs[math.ceil(npcId / TMT_EntriesPerFile)]

            if npcsArray then
                unitName = TMT_FindNameFromId(npcId, npcsArray)
            end
        end
    end

    return unitName
end

local function TMT_GetTooltipLineColor(line)
    -- In case the color was set in line directly in the text (combat pets for example)
    local color = string.match(_G["GameTooltipTextLeft" .. line]:GetText(), "|cff......")

    -- In case the color was set with the SetColor function
    if not color then 
        local r, g, b = _G["GameTooltipTextLeft" .. line]:GetTextColor()

        --[[
            Found somewhere on the internet that we had to adapt a bit the
            colors retrieved from the tooltip, but there was no explanation.
            It works though.
        ]]
        r = floor(r * 255 * 100 + 0.5) / 100
        g = floor(g * 255 * 100 + 0.5) / 100
        b = floor(b * 255 * 100 + 0.5) / 100
        color = TMT_RgbToHex({r, g, b}) 
    end 

    return color
end

local function TMT_GetNpcNameColor(unit)
    local color
    
    if TMT_NPCTooltip_UseBlizzardsColor then
        if TMT_NPCTooltip_DisplayNameInTitle then
            color = string.match(_G["GameTooltipTextLeft1"]:GetText(), "|cff......")

            if not color then
                color = "|r"
            end
        else
            color = TMT_GetTooltipLineColor(1)
        end
    else
        color = TMT_RgbToHex(TMT_NpcNameColor)
    end

    return color
end

--[[
    Talents are spells. But Blizzard gave these talents specific talents IDs that
    are not equal to the corresponding spell ID. And talents specific functions
    use the talent ID, not the spell ID.
    TMT only stores spells IDs. So when the addon is loaded, we use this function
    to build an array storing pairs of the form : {spell ID, talent ID}
]]
local function TMT_AssociateSpellsToTalents()
    TMT_Talents = nil
    TMT_Talents = {}
    local pair = {}

    for i = 0, TMT_SpellsMaxId do
        local talentId, name, texture, selected, available, spellId, unknowna, row, column, unknownb, known = GetTalentInfoByID(i)
        if spellId then
            pair = {talentId, spellId}
            table.insert(TMT_Talents, pair)
        end
    end
end

-- Same as just above but for PVP talents
local function TMT_AssociateSpellsToPvpTalents()
    TMT_PvpTalents = nil
    TMT_PvpTalents = {}
    local pair = {}

    for i = 0, TMT_SpellsMaxId do
        talentId, name, icon, selected, available, spellId, unlocked, row, column, known, unknown = GetPvpTalentInfoByID(i)
        if spellId then
            pair = {talentId, spellId}
            table.insert(TMT_PvpTalents, pair)
        end
    end
end

local function TMT_IsTalent(spellId)
    local isTalent = false

    for _, pair in ipairs(TMT_Talents) do
        if pair[2] == spellId then
            isTalent = true
            break
        end
    end
    
    return isTalent
end

local function TMT_IsPvpTalent(spellId)
    local isTalent = false

    for _, pair in ipairs(TMT_PvpTalents) do
        if pair[2] == spellId then
            isTalent = true
            break
        end
    end
    
    return isTalent
end

local function TMT_GetSpellNameColor()
    local color

    if TMT_SpellTooltip_UseClassColor then
        color = TMT_GetClassColor()
    else
        color = TMT_RgbToHex(TMT_SpellNameColor)
    end
    
    return color
end

local function TMT_FindSpellOriginalName(spellId)
    local spellName

    if spellId then
        local spellsArray = TMT.spells[math.ceil(spellId / TMT_EntriesPerFile)]

        if spellsArray then
            spellName = TMT_FindNameFromId(spellId, spellsArray)
        end
    end

    return spellName
end

local function TMT_GetItemNameColor(itemLink)
    local color

    if itemLink then
        color = string.match(itemLink, "(|c.+)|H")

        if not TMT_ItemTooltip_UseBlizzardsColor then
            color = TMT_RgbToHex(TMT_ItemNameColor)
        end    
    end

    return color
end

local function TMT_FindItemOriginalName(itemLink, tooltip)
    local itemName

    -- item link of the form "item:7073:0:0:0:0:0:0:0:0:0:0:0:0". 7073 is the ID here
    if itemLink then
        local itemID = tonumber(string.match(itemLink, "item:(%d+)"))

        if itemID then
            local itemsArray = TMT.items[math.ceil(itemID / TMT_EntriesPerFile)]

            if itemsArray then
                itemName = TMT_FindNameFromId(itemID, itemsArray)
            end
        end
    end

    return itemName
end

local function TMT_GetQuestNameColor(questId, isTask)
    local questTagInfos = C_QuestLog.GetQuestTagInfo(questId)
    local color, tagName, tagId, worldQuestType, quality, isElite

    if questTagInfos then
        tagName = questTagInfos["tagName"]
        tagId = questTagInfos["tagID"]
        worldQuestType = questTagInfos["worldQuestType"]
        quality = questTagInfos["quality"]
        isElite = questTagInfos["isElite"]

        --[[ debug 
        print("-- " .. questId .. " --")
        
        if tagId then
            print("tagId  : " .. tagId)
            print("Tag name : " .. tagName)
        end

        if worldQuestType then
            print("type   : " .. worldQuestType)
        end
        
        if quality then
            print("quality : " .. quality)
        end

        if isElite then
            print("isElite : yes")
        else
            print("isElite : no")
        end
        print("-----------------------")
        ]]
    end

    if isTask then -- world quests,  e.t.c.
        if tagId == 261 then -- Island expedition (BFA)
            color = TMT_Colors.yellow
        elseif tagId == 268 then -- Threat progression (BFA)
            color = TMT_Colors.white
        elseif quality then
            color = TMT_Colors.worldQuestQuality[quality + 1]
        else
            color = TMT_Colors.yellow
        end
    else -- normal quest
        color = TMT_Colors.yellow
    end

    if not TMT_QuestsNames_UseBlizzardsColor then
        color = TMT_RgbToHex(TMT_QuestNameColor)
    end

    return color
end

local function TMT_FindQuestOriginalName(questId)
    local questName

    if questId then
        local questsArray = TMT.quests[math.ceil(questId / TMT_EntriesPerFile)]

        if questsArray then
            questName = TMT_FindNameFromId(questId, questsArray)
        end
    end

    return questName
end

local function TMT_GetMinimapAreaId(zoneType)
    local currentAreaName
    local currentAreaId

    if zoneType == TMT_ZoneType.ZONE then
        currentAreaName = GetZoneText()
    elseif zoneType == TMT_ZoneType.SUBZONE then
        currentAreaName = GetMinimapZoneText()
    end

    for i = 1, TMT_AreasMaxId do
        local areaName = C_Map.GetAreaInfo(i)

        if areaName == currentAreaName then
            currentAreaId = i
            break;
        end
    end

    return currentAreaId
end

local function TMT_FindAreaOriginalName(areaId)
    local areaName

    if areaId then
        local areasArray = TMT.areas[math.ceil(areaId / TMT_EntriesPerFile)]
        areaName = TMT_FindNameFromId(areaId, areasArray)
    end

    return areaName
end

local function TMT_EntryPoint(isOn, task, param1, param2, param3)
    if isOn then
        if task == "quest" then -- param1 : quest id, param2 : isTask, param3 : tooltip
            local name = TMT_FindQuestOriginalName(param1)
            local color = TMT_GetQuestNameColor(param1, param2)
            TMT_TooltipDisplay(param3, name, TMT_QuestTooltip_DisplayNameInTitle, color)
        elseif task == "spell" then -- param1 : spell id, param2 : tooltip
            local name = TMT_FindSpellOriginalName(param1)

            if not TMT_IsNameAlreadyInTooltip(param2, name) then
                local color = TMT_GetSpellNameColor()
                TMT_TooltipDisplay(param2, name, TMT_SpellTooltip_DisplayNameInTitle, color)
            end
        elseif task == "talent" then -- param1 : spell id, param2 : tooltip
            local name = TMT_FindSpellOriginalName(param1)

            if not TMT_IsNameAlreadyInTooltip(param2, name) then                
                local color = TMT_GetSpellNameColor()
                TMT_TooltipDisplay(param2, name, TMT_SpellTooltip_DisplayNameInTitle, color)
            end
        elseif task == "npc" then -- param1 : unit, param2 : from guid, param3: tooltip
            local name = TMT_FindNpcOriginalName(param1, param2)
            local color = TMT_GetNpcNameColor(param1)
            TMT_TooltipDisplay(param3, name, TMT_NPCTooltip_DisplayNameInTitle, color)
        elseif task == "item" then -- param1 : item link, param2 : tooltip
            local name = TMT_FindItemOriginalName(param1, param2)
            
            if not TMT_IsNameAlreadyInTooltip(param2, name) then
                local color = TMT_GetItemNameColor(param1)
                TMT_TooltipDisplay(param2, name, TMT_ItemTooltip_DisplayNameInTitle, color)
            end
        elseif task == "area" then --param1 : tooltip
            local name
            local color = "|r"

            local zoneId = TMT_GetMinimapAreaId(TMT_ZoneType.ZONE)
            local subzoneId = TMT_GetMinimapAreaId(TMT_ZoneType.SUBZONE)

            local zoneName = TMT_FindAreaOriginalName(zoneId)
            local subzoneName =  TMT_FindAreaOriginalName(subzoneId)

            if zoneName and not (zoneName == GetZoneText()) then
                color = TMT_Colors.white
                name = zoneName

                if subzoneName and not (subzoneName == GetMinimapZoneText()) then
                    name = name .. "|r|n" .. subzoneName
                end
            elseif subzoneName and not (subzoneName == GetMinimapZoneText()) then
                name = subzoneName
            end
            
            TMT_TooltipDisplay(param1, name, true, color, true)
            GameTooltip:Show() 
        end
    end
end

--[[
    A few words on setUnitAura, setUnitBuff, etc.
    - setUnitAura is to work on the buffs AND debuffs on YOUR character
    - setUnitBuff is to work ONLY on the buffs on your target (can be yourself)
    - setUnitDebuff is to work ONLY on the debuffs on your target (can be yourself)
    - UnitAura retrieves ONLY buffs on your character and target
    - UnitBuff retrieves buffs on you and your target
    - UnitDebuff retrieves debuffs on you and your target

    Edit :
    My understanding for setUnitAura is only right on default Blizzard behavior. But
    it seems that it can be called on target with other addons, like Shadowed Unit Frame.
]]

-- Securehook handler. Called only on player with Blizzard. Called also on target with ShadowedUF and LunaUF
local function TMT_FindAuraID(casterID, unitID, auraIndex, filter)
        local actualSpellId

        -- The NOT_PASSIVE version was added in patch 1.13.3 for buffs and debuffs on yourself (but not yourself from target)
        if filter == "HELPFUL" or filter == "HELPFUL_NOT_PASSIVE" then
            local name, rank, count, debuffType, duration, expirationTime, unitCaster, isStealable, 
            shouldConsolidate, spellId = UnitBuff(unitID, auraIndex)
            
            actualSpellId = spellId
        elseif filter == "HARMFUL" or filter == "HARMFUL_NOT_PASSIVE" then
            local name, rank, count, debuffType, duration, expirationTime, unitCaster, isStealable, 
            shouldConsolidate, spellId = UnitDebuff(unitID, auraIndex)
    
            actualSpellId = spellId
        end
    
        TMT_EntryPoint(TMT_DisplaySpellsNames, "spell", actualSpellId, GameTooltip)
         -- since it's a secure hook, GameTooltip was already set. Have to show it again to adapt the size to the new text
        GameTooltip:Show()
end

-- Securehook handler
local function TMT_FindTargetBuffID(casterID, unitID, buffIndex)
    local name, rank, count, debuffType, duration, expirationTime, unitCaster, isStealable, 
    shouldConsolidate, spellId = UnitAura("TARGET", buffIndex)
        
    TMT_EntryPoint(TMT_DisplaySpellsNames, "spell", spellId, GameTooltip)
    GameTooltip:Show()
end

-- Securehook handler
local function TMT_FindTargetDebuffID(casterID, unitID, buffIndex)
    local name, rank, count, debuffType, duration, expirationTime, unitCaster, isStealable, 
    shouldConsolidate, spellId = UnitDebuff("TARGET", buffIndex)
  
    TMT_EntryPoint(TMT_DisplaySpellsNames, "spell", spellId, GameTooltip)
    GameTooltip:Show()
end

local function TMT_TalentTooltipSecureHook(gameTooltip, talentId, unknownBoolValue)
    local spellId = TMT_FindNameFromId(talentId, TMT_Talents)
    TMT_EntryPoint(TMT_DisplaySpellsNames, "talent", spellId, GameTooltip)
    GameTooltip:Show()
end

local function TMT_PvpTalentTooltipSecureHook(gameTooltip, talentId, unknownBoolValue)
    local spellId = TMT_FindNameFromId(talentId, TMT_PvpTalents)
    TMT_EntryPoint(TMT_DisplaySpellsNames, "talent", spellId, GameTooltip)
    GameTooltip:Show()
end

local function TMT_TaskPOIOnEnterSecureHook(TaskPOI)
    TMT_EntryPoint(TMT_DisplayQuestsNames, "quest", TaskPOI.questID, true, GameTooltip)
    GameTooltip:Show()
end

local function TMT_QuestMapLogTitleButtonOnEnterSecureHook(QuestMapLogTitle)
    TMT_EntryPoint(TMT_DisplayQuestsNames, "quest", QuestMapLogTitle.questID, false, GameTooltip)
    GameTooltip:Show()
end

local function TMT_ReagentTooltipSecureHook(unknown, recipeId, regeantIndex)
    local itemLink = C_TradeSkillUI.GetRecipeReagentItemLink(recipeId, regeantIndex)
    TMT_EntryPoint(TMT_DisplayItemsNames, "item", itemLink, GameTooltip)
    GameTooltip:Show()
end

local function TMT_ToyTooltipSecureHook(unknown, itemId)
    local itemLink = C_ToyBox.GetToyLink(itemId)
    TMT_EntryPoint(TMT_DisplayItemsNames, "item", itemLink, GameTooltip)
    GameTooltip:Show()
end

local function TMT_PetTooltipSecureHook(tooltip, petLink)
    speciesID, customName, level, xp, maxXp, displayID, isFavorite, name, icon, petType, creatureID, sourceText, description, isWild, canBattle, tradable, unique, obtainable = C_PetJournal.GetPetInfoByPetID(petLink)
    TMT_EntryPoint(TMT_DisplayNPCsNames, "npc", creatureID, false, tooltip)
end

local function TMT_SetHooks()
    -- Hooks to tooltip functions
    GameTooltip:HookScript("OnTooltipSetUnit", function(self) TMT_EntryPoint(TMT_DisplayNPCsNames, "npc", select(2, self:GetUnit()), true, self); end)
    GameTooltip:HookScript("OnTooltipSetSpell", function(self) TMT_EntryPoint(TMT_DisplaySpellsNames, "spell", select(2, self:GetSpell()), self); end)
    GameTooltip:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end)
    ItemRefTooltip:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end) -- tooltip when an item link is clicked
    ShoppingTooltip1:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end) -- tooltip when comparing items
    ShoppingTooltip2:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end) 
    ItemRefShoppingTooltip1:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end) -- tooltip when comparing items but from an item link
    ItemRefShoppingTooltip2:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end)
    hooksecurefunc(GameTooltip, "SetUnitAura", TMT_FindAuraID) -- Called on your character
    hooksecurefunc(GameTooltip, "SetUnitBuff", TMT_FindTargetBuffID) -- Only called on target (can be yourself)
    hooksecurefunc(GameTooltip, "SetUnitDebuff", TMT_FindTargetDebuffID) -- Only called on target (can be yourself)
    hooksecurefunc(GameTooltip, "SetTalent", TMT_TalentTooltipSecureHook) -- Called when hovering above a talent in the talents frame
    hooksecurefunc(GameTooltip, "SetPvpTalent", TMT_PvpTalentTooltipSecureHook) -- Called when hovering above a pvp talent in the pvp part of the talents frame
    hooksecurefunc(GameTooltip, "SetRecipeReagentItem", TMT_ReagentTooltipSecureHook)
    hooksecurefunc(GameTooltip, "SetToyByItemID", TMT_ToyTooltipSecureHook)
    hooksecurefunc(GameTooltip, "SetCompanionPet", TMT_PetTooltipSecureHook)
    MinimapZoneTextButton:HookScript("OnEnter", function(self) TMT_EntryPoint(true, "area", GameTooltip); end)

    -- Hooks to quests related functions
    hooksecurefunc("TaskPOI_OnEnter", TMT_TaskPOIOnEnterSecureHook) -- Called when hovering above a world quest POI on the map
    hooksecurefunc("QuestMapLogTitleButton_OnEnter", TMT_QuestMapLogTitleButtonOnEnterSecureHook) -- Called when hovering above a quest title in the quest log
end

 -- An invisible frame to respond to events
local TMT_eventFrame = CreateFrame("FRAME");
TMT_eventFrame:RegisterEvent("ADDON_LOADED"); -- Fired when saved variables are loaded
TMT_eventFrame:RegisterEvent("PLAYER_LOGIN"); 
TMT_eventFrame:RegisterEvent("PLAYER_LOGOUT"); -- Fired when player about to log out

function TMT_eventFrame:OnEvent(event, addonName)
    if event == "ADDON_LOADED" and addonName == "TheMotherTongue" then
        TMT_AssociateSpellsToTalents()
        TMT_AssociateSpellsToPvpTalents()

        if TMT_DisplayNPCsNames == nil then
            print(TMT_Colors.orange .. "Type \"/tmt\" to go to TheMotherTongue addon settings") -- The addon has just been installed
            TMT_DisplayNPCsNames = TMT_Defaults.displayNPCsNames
        end
        if TMT_NPCTooltip_DisplayNameInTitle == nil then
            TMT_NPCTooltip_DisplayNameInTitle = TMT_Defaults.npcTooltip_DisplayNameInTitle
        end
        if TMT_NamesDump == nil then
            TMT_NamesDump = {}
        end
        if TMT_DisplayQuestsNames == nil then 
            TMT_DisplayQuestsNames = TMT_Defaults.displayQuestsNames
        end
        if TMT_DisplaySpellsNames == nil then 
            TMT_DisplaySpellsNames = TMT_Defaults.displaySpellsNames
        end
        if TMT_SpellTooltip_DisplayNameInTitle == nil then 
            TMT_SpellTooltip_DisplayNameInTitle = TMT_Defaults.spellTooltip_DisplayNameInTitle
        end
        if TMT_NpcNameColor == nil then
            TMT_NpcNameColor = TMT_Defaults.npcNameColor
        end
        if TMT_SpellNameColor == nil then
            TMT_SpellNameColor = TMT_Defaults.spellNameColor
        end
        if TMT_QuestNameColor == nil then
            TMT_QuestNameColor = TMT_Defaults.questNameColor
        end
        if TMT_DisplayItemsNames == nil then
            TMT_DisplayItemsNames = TMT_Defaults.displayItemsNames
        end
        if TMT_ItemTooltip_DisplayNameInTitle == nil then
            TMT_ItemTooltip_DisplayNameInTitle = TMT_Defaults.itemTooltip_DisplayNameInTitle
        end
        if TMT_ItemNameColor == nil then
            TMT_ItemNameColor = TMT_Defaults.itemNameColor
        end
        if TMT_ItemTooltip_UseBlizzardsColor == nil then
            TMT_ItemTooltip_UseBlizzardsColor = TMT_Defaults.itemTooltip_UseBlizzardsColor
        end
        if TMT_QuestsNames_UseBlizzardsColor == nil then
            TMT_QuestsNames_UseBlizzardsColor = TMT_Defaults.questsNames_UseBlizzardsColor
        end
        if TMT_SpellTooltip_UseClassColor == nil then
            TMT_SpellTooltip_UseClassColor = TMT_Defaults.spellTooltip_UseClassColor
        end
        if TMT_Tooltips_DisplayNameUnderTitle == nil then
            TMT_Tooltips_DisplayNameUnderTitle = TMT_Defaults.tooltips_DisplayNameUnderTitle
        end
        if TMT_QuestTooltip_DisplayNameInTitle == nil then
            TMT_QuestTooltip_DisplayNameInTitle = TMT_Defaults.questTooltip_DisplayNameInTitle
        end
        if TMT_NPCTooltip_UseBlizzardsColor == nil then
            TMT_NPCTooltip_UseBlizzardsColor = TMT_Defaults.npcTooltip_UseBlizzardsColor
        end
        if TMT_LastUpdateMessageVersion == nil then
            TMT_LastUpdateMessageVersion = TMT_Defaults.lastUpdateMessageVersion
        end
    elseif event == "PLAYER_LOGIN" then
        TMT_DetectLanguagesPacks()
        TMT_CreateConfigPanel()
        TMT_SetHooks()
        TMT_DisplayUpdateMessage()
    end
end
TMT_eventFrame:SetScript("OnEvent", TMT_eventFrame.OnEvent);